/******************************************************************************
 * Copyright (C) 2009. Filip Maric
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ******************************************************************************/
#ifndef __URM_INSTRUCTION_HPP__
#define __URM_INSTRUCTION_HPP__

#include <string>
#include <sstream>
#include <iostream>
using namespace std;

#include "urm.hpp"

class URMInstruction {
public:

  void executeAndTrace(URM& m) {
    execute(m);
    m.traceInstruction(*this);
  }

  virtual void execute(URM& m) = 0;
  virtual void print(ostream& ostr) const = 0;

  static URMInstruction* parse(const string& str);
};

class ZInstruction : public URMInstruction {
public:
  ZInstruction(unsigned k) : _k(k) {}

  virtual void execute(URM& m) {
    m.set_register(_k, 0);
    m.advance();
  }

  virtual void print(ostream& ostr) const {
    ostr << "Z " << _k << endl;
  }

  static URMInstruction* parse(const string& str) {
    istringstream sstr(str.substr(2));
    unsigned k;
    sstr >> k;
    return new ZInstruction(k);
  }

private:
  unsigned _k;
};

class SInstruction : public URMInstruction {
public:
  SInstruction(unsigned k) : _k(k) {}

  virtual void execute(URM& m) {
    m.set_register(_k, m.get_register(_k) + 1);
    m.advance();
  }

  virtual void print(ostream& ostr) const {
    ostr << "S " << _k << endl;
  }

  static URMInstruction* parse(const string& str) {
    istringstream sstr(str.substr(2));
    unsigned k;
    sstr >> k;
    return new SInstruction(k);
  }
private:
  unsigned _k;
};

class TInstruction : public URMInstruction {
public:
  TInstruction(unsigned from, unsigned to) : _from(from), _to(to) {}

  virtual void execute(URM& m) {
    m.set_register(_to, m.get_register(_from));
    m.advance();
  }

  virtual void print(ostream& ostr) const {
    ostr << "T " << _from << " " << _to << endl;
  }

  static URMInstruction* parse(const string& str) {
    istringstream sstr(str.substr(2));
    unsigned from, to;
    sstr >> from;  
    sstr >> to;
    return new TInstruction(from, to);
  }

private:
  unsigned _from, _to;
};

class IInstruction : public URMInstruction {
public:
  IInstruction(unsigned k, unsigned val) : _k(k), _val(val) {}

  virtual void execute(URM& m) {
    m.set_register(_k, _val);
    m.advance();
  }

  virtual void print(ostream& ostr) const {
    ostr << "I " << _k << " " << _val << endl;
  }

  static URMInstruction* parse(const string& str) {
    istringstream sstr(str.substr(2));
    unsigned k, val;
    sstr >> k;
    sstr >> val;
    return new IInstruction(k, val);
  }

private:
  unsigned _k, _val;
};


class JInstruction : public URMInstruction {
public:
  JInstruction(unsigned r1, unsigned r2, string label) 
    : _r1(r1), _r2(r2), _label(label) {}

  virtual void execute(URM& m) {
    if (m.get_register(_r1) == m.get_register(_r2))
      m.gotoLabel(_label);
    else
      m.advance();
  }

  virtual void print(ostream& ostr) const {
    ostr << "J " << _r1 << " " << _r2 << " " << _label << endl;
  }

  static URMInstruction* parse(const string& str) {
    istringstream sstr(str.substr(2));
    unsigned r1, r2;
    string label;
    sstr >> r1;  
    sstr >> r2;
    sstr >> label;
    return new JInstruction(r1, r2, label);
  }

private:
  unsigned _r1, _r2;
  string _label;
};

class CommentInstruction : public URMInstruction {
public:
  CommentInstruction(const string& comment) : _comment(comment) {}

  virtual void execute(URM& m) {
    cout << _comment << endl;
    m.advance();
  }

  virtual void print(ostream& ostr) const {
    ostr << "; " << _comment << endl;
  }
  
  static URMInstruction* parse(const string& str) {
    return new CommentInstruction(str.substr(2));
  }

private:
  string _comment;
};

#endif
