/******************************************************************************
 * Copyright (C) 2009. Filip Maric
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ******************************************************************************/
#ifndef __URM_HPP__
#define __URM_HPP__

#include <iostream>
#include <string>
#include <vector>
#include <map>
using namespace std;

class URMInstruction;
#include "urm-trace.hpp"

class URM {
public:
  URM(URMTrace* trace = new URMTracePrint(cout)) 
    : _currentInstruction(1), _trace(trace) {}

  ~URM();

  void execute();

  void advance() {
    _currentInstruction++;
  }

  void gotoLabel(const string& label) {
    if (_labels.find(label) == _labels.end())
      throw string("Unkonwn label: ") + label;
    _currentInstruction = _labels[label];
  }

  void addInstruction(URMInstruction* instruction) {
    _instructions.push_back(instruction);
  }

  unsigned numInstructions() const {
    return _instructions.size();
  }

  void set_register(unsigned k, unsigned val) {
    ensure_size(k);
    _registers[k - 1] = val;
  }

  unsigned get_register(unsigned k) const {
    return k <= _registers.size() ? _registers[k - 1] : 0;
  }

  URMInstruction* get_instruction(unsigned k) const {
    return k <= _instructions.size() ? _instructions[k - 1] : 0;
  }

  void printInstructions(ostream& ostr) const;

  void printState(ostream& ostr) const {
    for (vector<unsigned>::const_iterator i = _registers.begin();
	 i != _registers.end();
	 i++)
      ostr << *i << " ";
    ostr << endl;
  }

  void printCurrentInstruction(ostream& ostr) const {
    ostr << _currentInstruction << endl;
  }

  void traceInstruction(const URMInstruction& i) {
    _trace->traceInstruction(i, *this);
  }


  void parse(istream& in);

  void addLabel(const string& label) {
    _labels[label] = _instructions.size() + 1;
  }

private:
  // Ensure that the vector has at least k elements
  void ensure_size(unsigned k) {
    int old_size = _registers.size();

    if (k < old_size)
      return;

    _registers.resize(k);

    // fill in the default zeros
    for (int i = old_size; i < k; i++)
      _registers[i] = 0;
  }

  unsigned _currentInstruction;
  vector<URMInstruction*> _instructions;
  vector<unsigned> _registers;
  map<string, unsigned> _labels;

  URMTrace* _trace;
};
#endif
